/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd 
* Moil NT 0810
* Australia
* 
* This software is GPL. Please read the COPYING file.
* ==================================================================
*
***************************************************************************/
#define VERSION    "0.1"
#define TITLE \
  "LA51 version " VERSION ", Copyright (C) 1999 Jason Nunn <jsno@downunder.net.au>\n" \
  "This software is GPL. Please read the COPYING file in the distribution.\n\n"

#define LN_STR_LEN   256
#define SYMBOL_LEN   40

/** structs **********************/
typedef struct _tbin_buf_ll
{
  char                *buf;
  int                 size;
  int                 org;
  struct _tbin_buf_ll *next;
} tbin_buf_ll;

#define SYMBOL_LABEL 0x01
#define SYMBOL_EQU   0x02
#define SYMBOL_CODE  0x04
#define SYMBOL_DATA  0x08
#define SYMBOL_IDATA 0x10
#define SYMBOL_XDATA 0x20
#define SYMBOL_BIT   0x40
#define SYMBOL_ANY   0xff

#define SYMBOL_ANY_8 (SYMBOL_EQU | SYMBOL_DATA | SYMBOL_IDATA)
#define SYMBOL_ANY_16 (SYMBOL_LABEL | SYMBOL_EQU | SYMBOL_CODE | SYMBOL_XDATA)

typedef struct _tsymbols_ll
{
  int                type;
  char               name[SYMBOL_LEN + 1];
  int                value;
  struct _tsymbols_ll *next;
} tsymbols_ll;


typedef struct _tjmps_ll
{
  char             name[SYMBOL_LEN + 1];
  int              line_no;
  int              op;
  tbin_buf_ll      *bin_buf;
  int              bin_ptr;
  int              jmp_addr;
  int              type;
  struct _tjmps_ll *next;
} tjmps_ll;

typedef struct
{
  char *name;
  void (*proc)(char **);
} top_tab;

/** ops ***********************************/
#define OP_ACALL      0x11

#define OP_ADD_REG    0x28
#define OP_ADD_DIR    0x25
#define OP_ADD_PTR    0x26
#define OP_ADD_DATA   0x24

#define OP_ADDC_REG   0x38
#define OP_ADDC_DIR   0x35
#define OP_ADDC_PTR   0x36
#define OP_ADDC_DATA  0x34

#define OP_AJMP       0x1

#define OP_ANL_AREG   0x58
#define OP_ANL_ADIR   0x55
#define OP_ANL_APTR   0x56
#define OP_ANL_ADATA  0x54
#define OP_ANL_DIR1   0x52
#define OP_ANL_DIR2   0x53
#define OP_ANL_CBIT   0x82
#define OP_ANL_CNBIT  0xb0

#define OP_CJNE_ADIR  0xb5
#define OP_CJNE_ADATA 0xb4
#define OP_CJNE_REG   0xb8
#define OP_CJNE_PTR   0xb6

#define OP_CLR_A      0xe4
#define OP_CLR_C      0xc3
#define OP_CLR_BIT    0xc2

#define OP_CPL_A      0xf4
#define OP_CPL_C      0xb3
#define OP_CPL_BIT    0xb2

#define OP_DA         0xd4

#define OP_DEC_A      0x14
#define OP_DEC_REG    0x18
#define OP_DEC_DIR    0x15
#define OP_DEC_PTR    0x16

#define OP_DIV        0x84

#define OP_DJNZ_REG   0xd8
#define OP_DJNZ_DIR   0xd5

#define OP_INC_DPTR   0xa3
#define OP_INC_A      0x04
#define OP_INC_REG    0x08
#define OP_INC_DIR    0x05
#define OP_INC_PTR    0x07

#define OP_JB_BIT     0x20
#define OP_JBC_BIT    0x10
#define OP_JC         0x40

#define OP_JNB_BIT    0x30
#define OP_JNC        0x50
#define OP_JNZ        0x70
#define OP_JZ         0x60

#define OP_JMP_PTR    0x73
#define OP_LCALL      0x12
#define OP_LJMP       0x02

#define OP_MOV_A_REG    0xe8
#define OP_MOV_A_DIR    0xe5
#define OP_MOV_A_PTR    0xe6
#define OP_MOV_A_DATA   0x74
#define OP_MOV_REG_A    0xf8
#define OP_MOV_REG_DIR  0xa8
#define OP_MOV_REG_DATA 0x78
#define OP_MOV_DIR_A    0xf5
#define OP_MOV_DIR_REG  0x88
#define OP_MOV_DIR_DIR  0x85
#define OP_MOV_DIR_PTR  0x86
#define OP_MOV_DIR_DATA 0x75
#define OP_MOV_PTR_A    0xf6
#define OP_MOV_PTR_DIR  0xa6
#define OP_MOV_PTR_DATA 0x76
#define OP_MOV_C_BIT    0xa2
#define OP_MOV_BIT_C    0x92
#define OP_MOV_DPTR     0x90

#define OP_MOVC_ADPTR   0x93
#define OP_MOVC_APC     0x83

#define OP_MOVX_A_PTR   0xe2
#define OP_MOVX_A_DPTR  0xe0
#define OP_MOVX_PTR_A   0xf2
#define OP_MOVX_DPTR_A  0xf0

#define OP_MUL        0xa4

#define OP_NOP        0x0

#define OP_ORL_AREG   0x48
#define OP_ORL_ADIR   0x45
#define OP_ORL_APTR   0x46
#define OP_ORL_ADATA  0x44
#define OP_ORL_DIR1   0x42
#define OP_ORL_DIR2   0x43
#define OP_ORL_CBIT   0x72
#define OP_ORL_CNBIT  0xa0

#define OP_POP        0xd0
#define OP_PUSH       0xc0

#define OP_RET        0x22
#define OP_RETI       0x32

#define OP_RL_A       0x23
#define OP_RLC_A      0x33
#define OP_RR_A       0x03
#define OP_RRC_A      0x13

#define OP_SETB_C     0xd3
#define OP_SETB_BIT   0xd2

#define OP_SJMP       0x80

#define OP_SUBB_REG   0x98
#define OP_SUBB_DIR   0x95
#define OP_SUBB_PTR   0x96
#define OP_SUBB_DATA  0x94

#define OP_SWAP_A     0xc4

#define OP_XCH_A_REG  0xc8
#define OP_XCH_A_DIR  0xc5
#define OP_XCH_A_PTR  0xc6

#define OP_XCHD_A_PTR 0xd6

#define OP_XRL_AREG   0x68
#define OP_XRL_ADIR   0x65
#define OP_XRL_APTR   0x66
#define OP_XRL_ADATA  0x64
#define OP_XRL_DIR1   0x62
#define OP_XRL_DIR2   0x63
