/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd
* Moil NT 0810
* Australia
*
* This software is GPL. Please read the COPYING file.
* ==================================================================
* expression processing
*
* nb/ the funny thing here is that it doesn't handle expressions haha ;)..
* but later on, full expression processing will be supported.
*
* right now, the code is only intelligent enough to evaluate ordinals and
* symbols
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "misc.h"
#include "symbols.h"

/***************************************************************************
*
***************************************************************************/
int get_string(char **s,char *str,int *len,int max_len)
{
  *len = 0;

  if(**s == '\'')
  {
    (*s)++;
    for(;;)
    {
      if(**s == 0)
      {
        p_error("unterminating string");
        return 0;
      }
      if(**s == '\'')
      {
        (*s)++;
        break;
      }
      if(*len >= max_len)
      {
        p_error("string too long");
        return 0;
      }
      str[*len] = **s;
      (*s)++;
      (*len)++;
    }
  }
  else
    return 0;
  return 1;
}

/***************************************************************************
*
***************************************************************************/
int get_single_hex(char *s,int term,int *val)
{
  term--;
  if(s[term] == 'h')
  {
    char ts[LN_STR_LEN + 1];

    strncpy(ts,s,term);
    ts[term] = 0;
    if((ts[0] >= 'a') && (ts[0] <= 'f'))
      p_warning("hex value starts with alpha");
    *val = 0;
    if(!sscanf(ts,"%x",val))
    {
      p_error(garbled_ord_mess);
      return 0;
    }
  }
  else
    return 0;
  return 1;
}

int get_single_bin(char *s,int term,int *val)
{
  term--;
  if(s[term] == 'b')
  {
    int neg = 0,pol = 1;
    char *ps;

    ps = s + term - 1;
    if(*s == '-')
    {
      neg = 1;
      s++;
    }
    *val = 0;
    for(;ps >= s;--ps)
    {
      if(*ps == '1')
        *val += pol;
      else
        if(*ps != '0')
        {
          p_error(garbled_ord_mess);
          return 0;
        }
      pol <<= 1;
    }
    if(neg)
      *val *= -1;
  }
  else
    return 0;
  return 1;
}

int get_single_dec(char *s,int term,int *val)
{
  int g(void)
  {
    char ts[LN_STR_LEN + 1];

    strncpy(ts,s,term);
    ts[term] = 0;
    *val = 0;
    if(!sscanf(ts,"%d",val))
    {
      p_error(garbled_ord_mess);
      return 0;
    }
    return 1;
  }

  term--;
  if(s[term] == 'd')
    g();
  else
    if((s[term] >= '0') && (s[term] <= '9'))
    {
      term++;
      g();
    }
    else
      return 0;
  return 1;
}

/***************************************************************************
*
***************************************************************************/
int get_single_ordinal(char **s,int *val)
{
  int stat = 1,term = 0;

  *val = 0;
  if(**s == '$')
  {
    (*s)++;
    if(bin_buf_curr != NULL)
    {
      *val = bin_buf_curr->org + bin_buf_curr->size;
      return 1;
    }
    return 0;
  }
  if(is_ordinal(**s))
  {
    for(;;)
    {
      char c;

      c = *(*s + term);
      if((c == ' ') || (c == ',') || (c == 0))
        break;
      term++;
    }
    if(term == 0)
    {
      p_error("Expecting ordinal");
      stat = 0;
    }
    else
    {
      if(!get_single_hex(*s,term,val))
        if(!get_single_bin(*s,term,val))
          if(!get_single_dec(*s,term,val))
          {
            p_error("unknown value base");
            stat = 0;
          }
    }
  }
  else
    stat = 0;
  if(stat)
    *s += term;
  return stat;
}

/***************************************************************************
*
***************************************************************************/
int get_16(char **s,int *val,int expected_sym,int op)
{
  int stat = 0,t = 0,len;
  char wstr[3] = "";

  if(!strncmp(*s,"high",4))
  {
    expected_sym |= SYMBOL_ANY_16;
    (*s) += 4;
    expect_ch(s,'(');
    skip_space(s);
    t = 2;
  }
  else if(!strncmp(*s,"low",3))
  {
    expected_sym |= SYMBOL_ANY_16;
    (*s) += 3;
    expect_ch(s,'(');
    skip_space(s);
    t = 1;
  }

  if(get_string(s,wstr,&len,2))
  {
    if(len == 1)
      *val = wstr[0];
    else if(len == 2)
    {
      *val = (wstr[0] << 8) + wstr[1];
    }
    else
      *val = 0;
    stat = 1;
  }
  else
    if(get_symbol_val(s,val,expected_sym,op,t))
      stat = 1;
    else
      if(get_single_ordinal(s,val))
        stat = 1;
  if(stat)
  {
    if((*val < -0x7fff) || (*val > 0xffff))
    {
      p_error(out_of_range_mess);
      return 0;
    }
    if(t)
    {
      skip_space(s);
      expect_ch(s,')');
      if(t == 2)
        *val >>= 8;
      else if(t == 1)
        *val &= 0xff;
    }
  }
  return stat;
}

int get_u16(char **s,int *val,int expected_sym,int op)
{
  int stat;

  stat = get_16(s,val,expected_sym,op);
  if(stat)
    if((*val < 0) || (*val > 0xffff))
    {
      p_error(out_of_range_mess);
      return 0;
    }
  return stat;
}

int get_8(char **s,int *val,int expected_sym,int op)
{
  int stat = 0;

  stat = get_16(s,val,expected_sym,op);
  if(stat)
  {
    if((*val < -128) || (*val > 256))
    {
      p_error(out_of_range_mess);
      return 0;
    }
  }
  return stat;
}
