/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd
* Moil NT 0810
* Australia
*
* This software is GPL. Please read the COPYING file.
* ==================================================================
* main
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "defs.h"
#include "misc.h"
#include "expressions.h"
#include "symbols.h"
#include "ops.h"

/***************************************************************************
*
***************************************************************************/
void assemble(void)
{
  while(!at_end)
  {
    char strln[LN_STR_LEN + 1];

    if(feof(fp))
      break;
    if(fgets(strln,LN_STR_LEN,fp) != NULL)
    {
      line_no++;
      if(strln[0] != ';')
      {
        char *s;
        int x,quote = 0;

        for(x = 0;x < strlen(strln);x++)
        {
          switch(strln[x])
          {
            case '\'':
              quote ^= 1;
              break;
            case '\t':
            case '\n':
            case '\r':
              strln[x] = ' ';
              break;
            case ';':
              if(!quote)
                strln[x] = 0;  /*ignores ";"'s in ('') strings*/
              break;
            default:
              if(!quote)
                strln[x] = tolower(strln[x]);
              break;
          }
        }
        s = strln;
        get_label(&s);
        get_op(&s);
        skip_space(&s);
        if(strlen(s))
          p_error(garbage_mess);
      }
    }
  }
}

int parse_file(char *fn)
{
  line_no = 0;
  filename = fn;
  fp = fopen(filename,"r");
  if(fp != NULL)
  {
    assemble();
    fclose(fp);
  }
  else
  {
    printf("Error: Couldn't open %s\n",filename);
    return 0;
  }
  return 1;
}

/***************************************************************************
*
***************************************************************************/
void print_summary(void)
{
  if(!at_end)
  {
    error_count++;
    printf("%s: error: EOF reached before end statement.\n",filename);
  }
  printf(
    "\n"
    "Error_count   : %d\n"
    "No of Warnings: %d\n"
    "Total Lines   : %d\n"
    "\n",error_count,warning_count,line_no);
}

void debug_summary(void)
{
  tjmps_ll   *j;
  tsymbols_ll *l;

  printf("\n==SYMBOLS =========\n");
  l = symbols_ll;
  while(l != NULL)
  {
    printf("%-40s %5d %4xh\n",
      l->name,l->type,l->value);
    l = l->next;
  }
  
  printf("\n==Jumps ==========\n");
  j = jmps_ll;
  while(j != NULL)
  {
    printf("%-40s %5d %4xh %4xh\n",
      j->name,j->line_no,j->bin_ptr,j->jmp_addr);
    j = j->next;
  }
}

/***************************************************************************
*
***************************************************************************/
#define HEX_LINE_LIM 0x10

void write_ihex(void)
{
  tbin_buf_ll *bp;
  int csum;

  void phex(int c)
  {
    char ct[] = "0123456789ABCDEF";

    c &= 0xff;
    putc(ct[c >> 4],fp);
    putc(ct[c & 0xf],fp);
  }

  void cphex(int c)
  {
    phex(c);
    csum += c;
  }

  bp = bin_buf_ll;
  while(bp != NULL)
  {
    int ptr = 0;

    for(;;)
    {
      int x,s;

      csum = 0;
      if(ptr == bp->size)
        break;
      else
      {
        s = bp->size - ptr;
        if(s > HEX_LINE_LIM)
          s = HEX_LINE_LIM;
      }
      putc(':',fp);
      cphex(s);
      x = bp->org + ptr;

      cphex(x >> 8);
      cphex(x & 0xff);
      cphex(0);
      for(x = 0;x < s;x++)
      {
        cphex(bp->buf[ptr]);
        ptr++;
      }
      phex(-csum & 0xff);
      putc('\n',fp);
    }
    bp = bp->next;
  }
  fprintf(fp,":00000001FF\n");
}

void save_ihex(char *fn)
{
  char str[LN_STR_LEN + 1] = "";
  char *p;

  strcpy(str,fn);
  p = strstr(str,".a51");
  if(p != NULL)
    strcpy(p,".hex");
  else
    strcat(str,".a51");  

  fp = fopen(str,"w");
  if(fp != NULL)
  {
    write_ihex();
    fclose(fp);
  }
  else
    printf("Error: couldn't create hex file.\n");
}

/***************************************************************************
*
***************************************************************************/
void pusage(void)
{
  printf("Usage: la51 -d <file.a51>\n\n");
}

int main(int argc,char *argv[])
{
  char *fn = NULL;

  printf(TITLE);
  if(argc > 1)
  {
    int x,debug_sum = 0;

    for(x = 0;x < argc;x++)
    {
      if(argv[x][0] == '-')
      {
        switch(argv[x][1])
        {
          case 'h':
            pusage();
            return 1;
          case 'd':
            debug_sum = 1;
            break;
        }
      }
      else
        fn = argv[x];
    }
    if(fn == NULL)
    {
      printf("No file specified.\n");
      return -1;
    }
    if(parse_file(PROG_PATH "/intrinsic.inc"))
    {
      if(parse_file(fn))
      {
        if(error_count == 0)
        {
          wirein_jmps();
          save_ihex(fn);
        }
        print_summary();
        if(debug_sum)
          debug_summary();
      }
    }
  }
  else
    pusage();

  return 1;
}
