/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd
* Moil NT 0810
* Australia
*
* This software is GPL. Please read the COPYING file.
* ==================================================================
* misc code
*
***************************************************************************/
#include <stdio.h>
#include <malloc.h>
#include <bytesex.h>
#include "defs.h"

FILE *fp;
char *filename;
int  at_end        = 0;
int  org_defined   = 0;
int  line_no       = 0;
int  org           = 0;
int  error_count   = 0;
int  warning_count = 0;

tbin_buf_ll *bin_buf_ll = NULL,**bin_buf_ll_tail = &bin_buf_ll,
  *bin_buf_curr = NULL;

char *garbled_ord_mess  = "garbled ordinal";
char *lab_too_long_mess = "label tooo lonng";
char *un_opand_mess     = "Unknown or illegal op-and";
char *un_opor_mess      = "Unknown op-or";
char *garbage_mess      = "trailing garbage";
char *out_of_range_mess = "ordinal out of range";
char *invalid_reg_mess  = "Invalid register";

/***************************************************************************
*
***************************************************************************/
void fatal_memory_error(void)
{
  printf("Error allocating memory. Aborting..\n");
  exit(0);
}

/***************************************************************************
*
***************************************************************************/
void p_error(char *str)
{
  error_count++;
  printf("%s:%d: error: %s.\n",filename,line_no,str);
}

void p_warning(char *str)
{
  warning_count++;
  printf("%s:%d: warning: %s.\n",filename,line_no,str);
}

void skip_space(char **s)
{
  for(;;)
  {
    if(**s != ' ')
      return;
    if(**s == 0)
      return;
    (*s)++;
  }
}

void expect_ch(char **s,char c)
{
  skip_space(s);
  if(**s != c)
  {
    char str[] = "Expecting ' '";

    str[strlen(str) - 2] = c;
    p_error(str);
  }
  else
    (*s)++;
}

int is_symbol(char c)
{
  if(((c >= 'a') && (c <= 'z')) || (c == '_'))
    return 1;
  return 0;
}

int is_ordinal(char c)
{
  if(((c >= '0') && (c <= '9')) || (c == '-'))
    return 1;
  return 0;
}

/***************************************************************************
*
***************************************************************************/
void append_bin_buf(int org)
{
  *bin_buf_ll_tail = malloc(sizeof(tbin_buf_ll));
  if(*bin_buf_ll_tail != NULL)
  {
    (*bin_buf_ll_tail)->buf = NULL;
    (*bin_buf_ll_tail)->size = 0;
    (*bin_buf_ll_tail)->org = org;
    (*bin_buf_ll_tail)->next = NULL;
    bin_buf_curr = *bin_buf_ll_tail;
    bin_buf_ll_tail = &((*bin_buf_ll_tail)->next);
  }
  else
    fatal_memory_error();
}

/***************************************************************************
*
***************************************************************************/
void no_org(void)
{
  printf("%s: error: ORG not defined.\n",filename);
}

void add_text8(char data)
{
  if(bin_buf_curr == NULL)
  {
    no_org();
    exit(0);
  }
  bin_buf_curr->size++;
  bin_buf_curr->buf = realloc(bin_buf_curr->buf,bin_buf_curr->size);
  if(bin_buf_curr->buf == NULL)
    fatal_memory_error();
  bin_buf_curr->buf[bin_buf_curr->size - 1] = data;
}

void add_text16(char *data)
{
#if   __BYTE_ORDER == __LITTLE_ENDIAN
  add_text8(*(data + 1));
  add_text8(*data);
#elif __BYTE_ORDER == __END_ENDIAN
  add_text8(*data);
  add_text8(*(data + 1));
#endif
}

void ammend_text8(char *buf,int ptr,char data)
{
  if(bin_buf_curr == NULL)
  {
    no_org();
    exit(0);
  }
  if(buf != NULL)
    buf[ptr] = data;
}

void ammend_text16(char *buf,int ptr,char *data)
{
#if __BYTE_ORDER == __LITTLE_ENDIAN
  ammend_text8(buf,ptr,*(data + 1));
  ammend_text8(buf,ptr + 1,*(data + 0));
#elif __BYTE_ORDER == __END_ENDIAN
  ammend_text8(buf,ptr,*(data + 0));
  ammend_text8(buf,ptr + 1,*(data + 1));
#endif
}
