/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd
* Moil NT 0810
* Australia
*
* This software is GPL. Please read the COPYING file.
* ==================================================================
* op processing code
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include "defs.h"
#include "misc.h"
#include "symbols.h"
#include "expressions.h"

/***************************************************************************
****************************************************************************
* reserved words
****************************************************************************
***************************************************************************/

/***************************************************************************
* compiler directives are ignored
***************************************************************************/
void rw_comp_dir(char **str)
{
  p_warning("compiler directives not implemented");
  for(;;)
    if(**str == 0)
      return;
    else
      (*str)++;
}

/***************************************************************************
* end
***************************************************************************/
void rw_end(char **dummy)
{
  at_end = 1;
}

/***************************************************************************
* org <u16>
***************************************************************************/
void rw_org(char **str)
{
  int v;

  skip_space(str);
  if(get_u16(str,&v,SYMBOL_ANY_16,0))
    append_bin_buf(v);
}

/***************************************************************************
* <label> db <8>,'<text>',<8>,<8>
***************************************************************************/
int next_d_ord(char **str)
{
  for(;;)
  {
    if(**str == 0)
      return 0;
    if(**str == ',')
    {
      (*str)++;
      break;
    }
    (*str)++;
  }
  skip_space(str);
  return 1;
}

void rw_db(char **str)
{
  int get_inst(void)
  {
    char ts[LN_STR_LEN + 1];
    int val;

    if(get_string(str,ts,&val,LN_STR_LEN))
    {
      int x;

      for(x = 0;x < val;x++)
        add_text8(ts[x]);
      return 1;
    }
    else
      if(get_8(str,&val,SYMBOL_ANY_8 | SYMBOL_BIT,0))
      {
        add_text8(val); 
        return 1;
      }
    return 0;
  }

  skip_space(str);
  if(get_inst())
    for(;;)
    {
      if(next_d_ord(str))
      {
        if(!get_inst())
          break;
      }
      else
        break;
    }
}

/***************************************************************************
* <label> dw <16>,<16>,<16>,'XX','X'
***************************************************************************/
void rw_dw(char **str)
{
  int get_inst(void)
  {
    char ts[LN_STR_LEN + 1];
    int val;

    if(get_string(str,ts,&val,LN_STR_LEN))
    {
      if(val == 1)
      {
        add_text8(0);
        add_text8(*ts);
        return 1;
      }
      else if(val == 2)
      {
        add_text8(ts[0]);
        add_text8(ts[1]);
        return 1;
      }
      else
        p_error("string too long");
    }
    else
      if(get_16(str,&val,SYMBOL_ANY,0))
      {
        add_text16((char *)&val); 
        return 1;
      }
    return 0;
  }

  skip_space(str);
  if(get_inst())
    for(;;)
    {
      if(next_d_ord(str))
      {
        if(!get_inst())
          break;
      }
      else
        break;
    }
}

/***************************************************************************
* <label> ds <array size>
***************************************************************************/
void rw_ds(char **str)
{
  int sz;

  skip_space(str);
  if(get_u16(str,&sz,SYMBOL_EQU,0))
    append_bin_buf(bin_buf_curr->org + bin_buf_curr->size + sz);
}

/***************************************************************************
****************************************************************************
* op codes
****************************************************************************
***************************************************************************/

/***************************************************************************
* misc funcs
***************************************************************************/
int reg_lex(char **str,int *val)
{
  (*str)++;
  if((**str >= '0') && (**str <= '7'))
  {
    *val = **str - '0';
    (*str)++;
    return 1;
  }
  else
    p_error(invalid_reg_mess);
  return 0;
}

int ptr_lex(char **str,int *val)
{
  *val = 0;
  (*str)++;
  if(!strncmp(*str,"r0",2))
    (*str) += 2;
  else if(!strncmp(*str,"r1",2))
  {
    (*str) += 2;
    *val = 1;
  }
  else
  {
    p_error(invalid_reg_mess);
    return 0;
  }
  return 1;
}

/***************************************************************************
* acall <u11>
* acall <label>
***************************************************************************/
void op_acall_common(char **str,int op)
{
  skip_space(str);
  if(get_jmp(op,str,0))
  {
    add_text8(0);
    add_text8(0);
  }
}

void op_acall(char **str)
{
  op_acall_common(str,OP_ACALL);
}

/***************************************************************************
* add a,rN
* add a,<8>
* add a,@rI
* add a,#<8>
***************************************************************************/
void op_add_common(char **str,int op_reg,int op_dir,int op_ptr,int op_data)
{
  int val;

  expect_ch(str,'a');
  expect_ch(str,',');
  skip_space(str);
  if(get_8(str,&val,SYMBOL_DATA,0))
  {
    add_text8(op_dir);
    add_text8(val);
  }
  else
    switch(**str)
    {
      case 'r':  /*register*/
        if(reg_lex(str,&val))
          add_text8(op_reg | val);
        break;
      case '@':  /*pointer (indirect)*/
        if(ptr_lex(str,&val))
          add_text8(op_ptr | val);
        break;
      case '#': /*immediate*/
        (*str)++;
        if(get_8(str,&val,SYMBOL_ANY_8,op_data))
        {
          add_text8(op_data);
          add_text8(val);
        }
        break;
      default:
        p_error(un_opand_mess);
        break;
    }
}

void op_add(char **str)
{
  op_add_common(str,OP_ADD_REG,OP_ADD_DIR,OP_ADD_PTR,OP_ADD_DATA);
}

/***************************************************************************
* addc a,rN
* addc a,<8>
* addc a,@rI
* addc a,#<8>
***************************************************************************/
void op_addc(char **str)
{
  op_add_common(str,OP_ADDC_REG,OP_ADDC_DIR,OP_ADDC_PTR,OP_ADDC_DATA);
}

/***************************************************************************
* ajmp <u11>
* ajmp <label>
***************************************************************************/
void op_ajmp(char **str)
{
  op_acall_common(str,OP_AJMP);
}

/***************************************************************************
* anl a,rN
* anl a,<8>
* anl a,@rI
* anl a,#<8>
* anl <8>,a
* anl <8>,#<8>
* anl c,<8>
* anl c,/<8>
***************************************************************************/
void op_anl_common(
  char **str,
  int op_areg,int op_adir,int op_aptr,int op_adata,int op_dir1,int op_dir2,
  int op_cbit,int op_cnbit,int bit_op_f)
{
  int opand;

  skip_space(str);
  if(get_8(str,&opand,SYMBOL_DATA,0))
  {
    expect_ch(str,',');
    skip_space(str);
    switch(**str)
    {
      case 'a':
        (*str)++;
        add_text8(op_dir1);
        add_text8(opand);
        break;
      case '#':
      {
        int param;

        (*str)++;
        if(get_8(str,&param,SYMBOL_ANY_8,op_dir2))
        {
          add_text8(op_dir2);
          add_text8(opand);
          add_text8(param);
        }
        break;
      }
      default:
        p_error(un_opand_mess);
        break;
    }
  }
  else
    switch(**str)
    {
      case 'a':
        op_add_common(str,op_areg,op_adir,op_aptr,op_adata);
        break;
      case 'c':
        if(bit_op_f)
        {
          int op;

          (*str)++;
          expect_ch(str,',');
          skip_space(str);
          if(**str == '/')
          {
            op = op_cnbit;
            (*str)++;
          }
          else
            op = op_cbit;
          add_text8(op);
          if(get_8(str,&opand,SYMBOL_BIT,0))
            add_text8(opand);
          break;
        }
      default:
        p_error(un_opand_mess);
        break;
    }
}

void op_anl(char **str)
{
  op_anl_common(str,
    OP_ANL_AREG,OP_ANL_ADIR,OP_ANL_APTR,OP_ANL_ADATA,
    OP_ANL_DIR1,OP_ANL_DIR2,OP_ANL_CBIT,OP_ANL_CNBIT,1);
}

/***************************************************************************
* cjne a,<8>,<8>
* cjne a,#<8>,<8>
* cjne rN,#<8>,<8>
* cjne @ri,#<8>,<8>
***************************************************************************/
void op_cjne(char **str)
{
  int x;

  int get_rel(int op)
  {
    expect_ch(str,',');
    skip_space(str);
    if(get_jmp(op,str,0))
      return 1;
    return 0;
  }

  void get_dat(int sym_exp,int set_op,int op_or)
  {
    int v;

    if(get_8(str,&v,sym_exp,0))
      if(get_rel(set_op))
      {
        add_text8(set_op | op_or);
        add_text8(v);
        add_text8(0);
      }
  }

  skip_space(str);   
  switch(**str)
  {
    case 'a':
      (*str)++;
      expect_ch(str,',');
      skip_space(str);
      if(**str == '#')
      {
        (*str)++;
        get_dat(SYMBOL_ANY_8,OP_CJNE_ADATA,0);
      }
      else
        get_dat(SYMBOL_DATA,OP_CJNE_ADIR,0);
      break;
    case 'r':
      if(reg_lex(str,&x))
      {
        expect_ch(str,',');
        expect_ch(str,'#');
        get_dat(SYMBOL_ANY_8,OP_CJNE_REG,x);
      }
      break;
    case '@':
      if(ptr_lex(str,&x))
      {
        expect_ch(str,',');
        expect_ch(str,'#');
        get_dat(SYMBOL_ANY_8,OP_CJNE_PTR,x);
      }
      break;
    default:
      p_error(un_opand_mess);
      break;
  }
}

/***************************************************************************
* clr  a
* clr  c
* clr  [<8> | sym.2]
***************************************************************************/
void op_clr_common(char **str,int op_clr_a,int op_clr_c,int op_clr_bit)
{
  int opand;

  skip_space(str);
  if(get_8(str,&opand,SYMBOL_DATA | SYMBOL_BIT,0))
  {
    add_text8(op_clr_bit);
    add_text8(opand);
  }
  else
    switch(**str)
    {
      case 'a':
        (*str)++;
        add_text8(op_clr_a);
        break;
      case 'c':
        (*str)++;
        add_text8(op_clr_c);
        break;
      default:
        p_error(un_opand_mess);
        break;
    }
}

void op_clr(char **str)
{
  op_clr_common(str,OP_CLR_A,OP_CLR_C,OP_CLR_BIT);
}

/***************************************************************************
* cpl  a
* cpl  c
* cpl  [<8> | sym.2]
***************************************************************************/
void op_cpl(char **str)
{
  op_clr_common(str,OP_CPL_A,OP_CPL_C,OP_CPL_BIT);
}

/***************************************************************************
* da  a
***************************************************************************/
void op_da(char **str)
{
  skip_space(str);
  if(**str == 'a')
  {
    (*str)++;
    add_text8(OP_DA);
  }
  else
    p_error(un_opand_mess);
}

/***************************************************************************
* dec a
* dec rN
* dec <8>
* dec @rI
***************************************************************************/
void op_dec_common(char **str,int op_a,int op_reg,int op_dir,int op_ptr)
{
  int val;

  skip_space(str);
  if(get_8(str,&val,SYMBOL_DATA,0))
  {
    add_text8(op_dir);
    add_text8(val);
  }
  else
    switch(**str)
    {
      case 'a':  /*acc*/
        (*str)++;
        add_text8(op_a);
        break;
      case 'r':  /*register*/
        if(reg_lex(str,&val))
          add_text8(op_reg | val);
        break;
      case '@':  /*pointer (indirect)*/
        if(ptr_lex(str,&val))
          add_text8(op_ptr | val);
        break;
      default:
        p_error(un_opand_mess);
        break;
    }
}

void op_dec(char **str)
{
  op_dec_common(str,OP_DEC_A,OP_DEC_REG,OP_DEC_DIR,OP_DEC_PTR);
}

/***************************************************************************
* div ab
***************************************************************************/
void op_div_common(char **str,int op)
{
  skip_space(str);
  if(!strncmp(*str,"ab ",3))
  {
    (*str) += 2;
    add_text8(op);
  }
  else
    p_error(un_opand_mess);
}

void op_div(char **str)
{
  op_div_common(str,OP_DIV);
}

/***************************************************************************
* djnz rN,<8>
* djnz <8>,<8>
***************************************************************************/
void op_djnz(char **str)
{
  int val;

  skip_space(str);
  if(get_8(str,&val,SYMBOL_DATA,0))
  {
    expect_ch(str,',');
    skip_space(str);
    if(get_jmp(OP_DJNZ_DIR,str,0))
    {
      add_text8(OP_DJNZ_DIR);
      add_text8(val);
      add_text8(0);
    }
  }
  else
    if(**str == 'r')
    {
      if(reg_lex(str,&val))
      {
        expect_ch(str,',');
        skip_space(str);
        if(get_jmp(OP_DJNZ_REG,str,0))
        {
          add_text8(OP_DJNZ_REG | val);
          add_text8(0);
        }
      }
    }
}

/***************************************************************************
* inc dptr
* inc a
* inc rN
* inc <8>
* inc @rI
***************************************************************************/
void op_inc(char **str)
{
  skip_space(str);
  if(!strncmp(*str,"dptr ",5))
  {
    (*str) += 4;
    add_text8(OP_INC_DPTR);
  }
  else
    op_dec_common(str,OP_INC_A,OP_INC_REG,OP_INC_DIR,OP_INC_PTR);
}

/***************************************************************************
* jb <8>,<8>
***************************************************************************/
void op_jb_common(char **str,int op)
{
  int opand;

  skip_space(str);
  if(get_8(str,&opand,SYMBOL_DATA | SYMBOL_BIT,0))
  {
    expect_ch(str,',');
    skip_space(str);
    if(get_jmp(op,str,0))
    {
      add_text8(op);
      add_text8(opand);
      add_text8(0);
    }
  }
  else
    p_error(un_opand_mess);
}

void op_jb(char **str)
{
  op_jb_common(str,OP_JB_BIT);
}

/***************************************************************************
* jbc <8>,<8>
***************************************************************************/
void op_jbc(char **str)
{
  op_jb_common(str,OP_JBC_BIT);
}

/***************************************************************************
* jc <8>
***************************************************************************/
void op_jc_common(char **str,int op)
{
  skip_space(str);
  if(get_jmp(op,str,0))
  {
    add_text8(op);
    add_text8(0);
  }
}

void op_jc(char **str)
{
  op_jc_common(str,OP_JC);
}

/***************************************************************************
* jnb <8>,<8>
***************************************************************************/
void op_jnb(char **str)
{
  op_jb_common(str,OP_JNB_BIT);
}

/***************************************************************************
* jnc <8>
***************************************************************************/
void op_jnc(char **str)
{
  op_jc_common(str,OP_JNC);
}

/***************************************************************************
* jnz <8>
***************************************************************************/
void op_jnz(char **str)
{
  op_jc_common(str,OP_JNZ);
}

/***************************************************************************
* jz <8>
***************************************************************************/
void op_jz(char **str)
{
  op_jc_common(str,OP_JZ);
}

/***************************************************************************
* lcall <u16>
* lcall <label>
***************************************************************************/
void op_lcall_common(char **str,int op)
{
  skip_space(str);
  if(get_jmp(op,str,0))
  {
    add_text8(op);
    add_text8(0);
    add_text8(0);
  }
}

void op_lcall(char **str)
{
  op_lcall_common(str,OP_LCALL);
}

void op_call(char **str)
{
  op_lcall_common(str,OP_LCALL);
  p_warning("Assuming CALL to be LCALL");
}

/***************************************************************************
* jmp @a+dptr
***************************************************************************/
void op_jmp(char **str)
{
  skip_space(str);
  if(**str == '@')
  {
    if(!strncmp(*str,"@a+dptr ",8))
    {
      (*str) += 7;
      add_text8(OP_JMP_PTR);
    }
    else
      p_error(un_opand_mess);
  }
  else
  {
    op_lcall_common(str,OP_LCALL);
    p_warning("Assuming JMP to be an LJMP");
  }
}

/***************************************************************************
* ljmp <u16>
* ljmp <label>
***************************************************************************/
void op_ljmp(char **str)
{
  op_lcall_common(str,OP_LJMP);
}

/***************************************************************************
* mov a,rN
* mov a,<8>
* mov a,@rI
* mov a,#<8>
*
* mov c,<8>
*
* mov @rI,a
* mov @rI,<8>
* mov @rI,#<8>
*
* mov rN,a
* mov rN,<8>
* mov rN,#<8>
*
* mov dptr,#<16>
*
* mov <8>,a
* mov <8>,rN
* mov <8>,<8>
* mov <8>,@rI
* mov <8>,#<8>
* mov <8>,c
***************************************************************************/
void op_mov(char **str)
{
  int val;
  char c2,c3,c4,c5;

  void mov_c(void)
  {
    int v;

    (*str)++;
    expect_ch(str,',');
    skip_space(str);
    if(get_8(str,&v,SYMBOL_DATA | SYMBOL_BIT,0))
    {
      add_text8(OP_MOV_C_BIT);
      add_text8(v);
    }
  }

  void s_tpl(int i,int op_dir,int op_a,int op_data)
  {
    int v;

    expect_ch(str,',');
    skip_space(str);
    if(get_8(str,&v,SYMBOL_DATA,0))
    {
      add_text8(op_dir | i);
      add_text8(v);
    }
    else
      switch(**str)
      {
        case 'a':
          (*str)++;
          add_text8(op_a | i);
          break;
        case '#':
          (*str)++;
          if(get_8(str,&v,SYMBOL_ANY_8,op_data))
          {
            add_text8(op_data | i);
            add_text8(v);
          }
          break;
        default:  
          p_error(un_opand_mess);
      }
  }

  void mov_ptr(void)
  {
    int ptr;

    *str += 2;
    ptr = **str - '0';
    if((ptr == 0) || (ptr == 1))
    {
      (*str)++;
      s_tpl(ptr,OP_MOV_PTR_DIR,OP_MOV_PTR_A,OP_MOV_PTR_DATA);
    }
    else
      p_error(un_opand_mess);
  }

  void mov_reg(void)
  {
    int r;

    (*str)++;
    r = **str - '0';
    (*str)++;
    s_tpl(r,OP_MOV_REG_DIR,OP_MOV_REG_A,OP_MOV_REG_DATA);
  }

  void dir_op(void)
  {
    int v;

    if(get_8(str,&v,SYMBOL_DATA,0))
    {
      int v2;

      expect_ch(str,',');
      skip_space(str);
      if(get_8(str,&v2,SYMBOL_DATA,0))
      {
        add_text8(OP_MOV_DIR_DIR);
        add_text8(v2);
        add_text8(v);
      }
      else
        switch(**str)
        {
          case 'a':
            (*str)++;
            add_text8(OP_MOV_DIR_A);
            add_text8(v);
            break;
          case 'r':
            if(reg_lex(str,&v2))
            {
              add_text8(OP_MOV_DIR_REG | v2);
              add_text8(v);
            }
            break;
          case '@':
            if(ptr_lex(str,&v2))
            {
              add_text8(OP_MOV_DIR_PTR | v2);
              add_text8(v);
            }
            break;
          case '#':
            (*str)++;
            if(get_8(str,&v2,SYMBOL_ANY_8,OP_MOV_DIR_DATA))
            {
              add_text8(OP_MOV_DIR_DATA);
              add_text8(v);
              add_text8(v2);
            }
            break;
          default:
            p_error(un_opand_mess);
        }
    }
    else
      p_error(un_opand_mess);
  }

  skip_space(str);
  c2 = *(*str + 1);
  c3 = *(*str + 2);
  c4 = *(*str + 3);
  c5 = *(*str + 4);
  if((**str == 'a') && ((c2 == ' ') || (c2 == ',')))
    op_add_common(str,OP_MOV_A_REG,OP_MOV_A_DIR,OP_MOV_A_PTR,OP_MOV_A_DATA);
  else if((**str == 'c') && ((c2 == ' ') || (c2 == ',')))
    mov_c();
  else if((**str == '@') && (c2 == 'r') &&
         ((c4 == ' ') || (c4 == ',')))
    mov_ptr();
  else if((**str == 'r') && ((c2 >= '0') && (c2 <= '7')) &&
         ((c3 == ' ') || (c3 == ',')))
    mov_reg();
  else if(!strncmp(*str,"dptr",4) && ((c5 == ' ') || (c5 == ',')))
  {
    (*str) += 4;
    expect_ch(str,',');
    expect_ch(str,'#');
    if(get_u16(str,&val,SYMBOL_ANY_16,OP_MOV_DPTR))
    {
      add_text8(OP_MOV_DPTR);
      add_text16((char *)&val);
    }
    else
      p_error(un_opand_mess);
  }
  else
  {
    char *ts;

    ts = strchr(*str,',');
    if(ts != NULL)
    {
      ts++;
      skip_space(&ts);
      if((ts[0] == 'c') && (ts[1] == ' '))   /*is mov bit,c ?*/
      {
        if(get_8(str,&val,SYMBOL_DATA | SYMBOL_BIT,0))
        {
          expect_ch(str,',');
          expect_ch(str,'c');
          add_text8(OP_MOV_BIT_C);
          add_text8(val);
        }
        else
          p_error(un_opand_mess);
      }
      else  /*8 bit op*/
        dir_op();
    }
    else
      p_error(un_opand_mess);
  }
}

/***************************************************************************
* movc a,@a+dptr
* movc a,@a+pc
***************************************************************************/
void op_movc(char **str)
{
  skip_space(str);
  expect_ch(str,'a');
  expect_ch(str,',');
  skip_space(str);
  if(**str == '@')
  {
    if(!strncmp(*str,"@a+dptr ",8))
    {
      (*str) += 7;
      add_text8(OP_MOVC_ADPTR);
    }
    else if(!strncmp(*str,"@a+pc ",6))
    {
      (*str) += 5;
      add_text8(OP_MOVC_APC);
    }
    else
      p_error(un_opand_mess);
  }
  else
    p_error(un_opand_mess);
}

/***************************************************************************
* movx a,@rI
* movx a,@dptr
* movx @rI,a
* movx @dptr,a
***************************************************************************/
void op_movx(char **str)
{
  char c2;

  skip_space(str);
  c2 = *(*str + 1);
  if((**str == 'a') && ((c2 == ' ') || (c2 == ',')))
  {
    (*str)++;
    expect_ch(str,',');
    skip_space(str);
    if(**str == '@')
    {
      (*str)++;
      if(!strcmp(*str,"dptr "))
      {
        *str += 4;
        add_text8(OP_MOVX_A_DPTR);
      }
      else
      {
        int v;

        if(reg_lex(str,&v))
          add_text8(OP_MOVX_A_PTR | v);
      }
    }
    else
      p_error(un_opand_mess);
  }
  else if(**str == '@')
  {
    c2 = *(*str + 5);
    if(!strncmp(*str,"@dptr",5) && ((c2 == ' ') || (c2 == ',')))
    { 
      *str += 5;
      expect_ch(str,',');
      expect_ch(str,'a');
      add_text8(OP_MOVX_DPTR_A);
    }
    else
    {
      int v;

      if(ptr_lex(str,&v))
      {

        expect_ch(str,',');
        expect_ch(str,'a');
        add_text8(OP_MOVX_PTR_A | v);
      }
    }
  }
  else
    p_error(un_opand_mess);
}

/***************************************************************************
* mul ab
***************************************************************************/
void op_mul(char **str)
{
  op_div_common(str,OP_MUL);
}

/***************************************************************************
* nop
***************************************************************************/
void op_nop(char **dummy)
{
  add_text8(OP_NOP);
}

/***************************************************************************
* orl a,rN
* orl a,<8>   
* orl a,@rI
* orl a,#<8>
* orl <8>,a
* orl <8>,#<8>
* orl c,<8>
* orl c,/<8>
***************************************************************************/
void op_orl(char **str)
{
  op_anl_common(str,
    OP_ORL_AREG,OP_ORL_ADIR,OP_ORL_APTR,OP_ORL_ADATA,
    OP_ORL_DIR1,OP_ORL_DIR2,OP_ORL_CBIT,OP_ORL_CNBIT,1);
}

/***************************************************************************
* pop  <8>
* push <8>
***************************************************************************/
void op_pop_common(char **str,int op)
{
  int v;

  skip_space(str);
  if(get_8(str,&v,SYMBOL_DATA,0))
  {
    add_text8(op);
    add_text8(v);
  }
  else
    p_error(un_opand_mess);
}

void op_pop(char **str)
{
  op_pop_common(str,OP_POP);
}

void op_push(char **str)
{
  op_pop_common(str,OP_PUSH);
}

/***************************************************************************
* ret
* reti
***************************************************************************/
void op_ret(char **dummy)
{
  add_text8(OP_RET);
}

void op_reti(char **dummy)
{
  add_text8(OP_RETI);
}

/***************************************************************************
* rl  a
* rr  a
* rlc a
* rrc a
***************************************************************************/
void op_rl_common(char **str,int op)
{
  skip_space(str);
  expect_ch(str,'a');
  if(**str == ' ')
    add_text8(op);
  else
    p_error(un_opand_mess);
}

void op_rl(char **str)
{
  op_rl_common(str,OP_RL_A);
}

void op_rr(char **str)
{
  op_rl_common(str,OP_RR_A);
}

void op_rlc(char **str)
{
  op_rl_common(str,OP_RLC_A);
}

void op_rrc(char **str)
{
  op_rl_common(str,OP_RRC_A);
}

/***************************************************************************
* swap a
***************************************************************************/
void op_swap(char **str)
{
  op_rl_common(str,OP_SWAP_A);
}

/***************************************************************************
* setb c
* setb [<8> | sym.2]
***************************************************************************/
void op_setb(char **str)
{
  int opand;

  skip_space(str);
  if(get_8(str,&opand,SYMBOL_DATA | SYMBOL_BIT,0))
  {
    add_text8(OP_SETB_BIT);
    add_text8(opand);
  }
  else
    switch(**str)
    {
      case 'c':
        (*str)++;
        add_text8(OP_SETB_C);
        break;
      default:
        p_error(un_opand_mess);
        break;
    }
}

/***************************************************************************
* sjmp <8>
***************************************************************************/
void op_sjmp(char **str)
{
  op_jc_common(str,OP_SJMP);
}

/***************************************************************************
* subb a,rN
* subb a,<8>
* subb a,@rI
* subb a,#<8>
***************************************************************************/
void op_subb(char **str)
{
  op_add_common(str,OP_SUBB_REG,OP_SUBB_DIR,OP_SUBB_PTR,OP_SUBB_DATA);
}

/***************************************************************************
* xch a,rN
* xch a,<8>
* xch a,@rI
***************************************************************************/
void op_xch(char **str)
{
  int val;

  expect_ch(str,'a');
  expect_ch(str,',');
  skip_space(str);
  if(get_8(str,&val,SYMBOL_DATA,0))
  {
    add_text8(OP_XCH_A_DIR);
    add_text8(val);
  }
  else
    switch(**str)
    {
      case 'r':  /*register*/
        if(reg_lex(str,&val))
          add_text8(OP_XCH_A_REG | val);
        break;
      case '@':  /*pointer (indirect)*/
        if(ptr_lex(str,&val))
          add_text8(OP_XCH_A_PTR | val);
        break;
      default:
        p_error(un_opand_mess);
        break;
    }
}

/***************************************************************************
* xchd a,@rI
***************************************************************************/
void op_xchd(char **str)
{
  expect_ch(str,'a');
  expect_ch(str,',');   
  skip_space(str);
  if(**str == '@')
  {
    int val;

    if(ptr_lex(str,&val))
      add_text8(OP_XCHD_A_PTR | val);
  }
  else
    p_error(un_opand_mess);
}

/***************************************************************************
* xrl a,rN
* xrl a,<8>   
* xrl a,@rI
* xrl a,#<8>
* xrl <8>,a
* xrl <8>,#<8>
***************************************************************************/
void op_xrl(char **str)
{
  op_anl_common(str,
    OP_XRL_AREG,OP_XRL_ADIR,OP_XRL_APTR,OP_XRL_ADATA,
    OP_XRL_DIR1,OP_XRL_DIR2,0,0,0);
}

/***************************************************************************
****************************************************************************
*
****************************************************************************
***************************************************************************/
top_tab op_tab[] =
{
  {"end",  rw_end},
  {"org",  rw_org},
  {"db",   rw_db},
  {"dw",   rw_dw},
  {"ds",   rw_ds},

  {"acall",op_acall},
  {"add",  op_add},
  {"addc", op_addc},
  {"ajmp", op_ajmp},
  {"anl",  op_anl},
  {"call", op_call},
  {"cjne", op_cjne},
  {"clr",  op_clr},
  {"cpl",  op_cpl},
  {"da",   op_da},
  {"dec",  op_dec},
  {"div",  op_div},
  {"djnz", op_djnz},
  {"inc",  op_inc},
  {"jb",   op_jb},
  {"jbc",  op_jbc},
  {"jc",   op_jc},
  {"jnb",  op_jnb},
  {"jnc",  op_jnc},
  {"jnz",  op_jnz},
  {"jz",   op_jz},
  {"jmp",  op_jmp},
  {"lcall",op_lcall},
  {"ljmp", op_ljmp},
  {"mov",  op_mov},
  {"movc", op_movc},
  {"movx", op_movx},
  {"mul",  op_mul},
  {"nop",  op_nop},
  {"ret",  op_ret},
  {"reti", op_reti},
  {"pop",  op_pop},
  {"push", op_push},
  {"orl",  op_orl},
  {"rl",   op_rl},
  {"rr",   op_rr},
  {"rlc",  op_rlc},
  {"rrc",  op_rrc},
  {"setb", op_setb},
  {"sjmp", op_sjmp},
  {"subb", op_subb},
  {"swap", op_swap},
  {"xch",  op_xch},
  {"xchd", op_xchd},
  {"xrl",  op_xrl},
  {NULL, NULL}
};

void get_op(char **sl)
{
  int p;
  char ts[LN_STR_LEN + 1];

  skip_space(sl);
  if(**sl == 0)
    return;
  if(**sl == '$')
  {
    rw_comp_dir(sl);
    return;
  }
  if(sscanf(*sl,"%s",ts) == 1)
  {
    *sl += strlen(ts);
    for(p = 0;;p++)
    {
      if(op_tab[p].name == NULL)
      {
        get_def_symbol(sl,ts);
        return;
      }
      if(!strcmp(ts,op_tab[p].name))
      {
        op_tab[p].proc(sl);
        return;
      }
    }
  }
}
