/***************************************************************************
* la51- Intel 8051 Cross Assembler
* By Jason Nunn <jsno@downunder.net.au>
*
* 32 Rothdale Rd
* Moil NT 0810
* Australia
*
* This software is GPL. Please read the COPYING file.
* ==================================================================
* symbol processing related code
*
***************************************************************************/
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "defs.h"
#include "misc.h"
#include "expressions.h"

tsymbols_ll *symbols_ll = NULL;
tjmps_ll    *jmps_ll = NULL,**jmps_ll_tail = &jmps_ll;

/***************************************************************************
*
***************************************************************************/
void append_symbol(char *name,int type,int val)
{
  void f(tsymbols_ll **p)
  {
    if(*p != NULL)
    {
      if(!strcmp((*p)->name,name))
        p_error("label/symbol already defined");
      else
        f(&((*p)->next));
    }
    else
    {
      *p = malloc(sizeof(tsymbols_ll));
      if(*p != NULL)
      {
        (*p)->type = type;
        strncpy((*p)->name,name,SYMBOL_LEN);
        (*p)->name[SYMBOL_LEN] = 0;
        (*p)->value = val;
        (*p)->next = NULL;
      }
      else
        fatal_memory_error();
    }
  }

  if(strlen(name) > SYMBOL_LEN)
    p_error(lab_too_long_mess);
  else
    f(&symbols_ll);
}

tsymbols_ll *find_symbol(char *name)
{
  tsymbols_ll *s = NULL;

  void f(tsymbols_ll *p)
  {
    if(p != NULL)
    {
      if(!strcmp(p->name,name))
        s = p;
      else
        f(p->next);
    }
  }
  f(symbols_ll);
  return s;
}

/***************************************************************************
*
***************************************************************************/
void append_jmp(int op,char *name,int jmp_addr,int type)
{
  if(strlen(name) > SYMBOL_LEN)
    p_error(lab_too_long_mess);
  {
    *jmps_ll_tail = malloc(sizeof(tjmps_ll));
    if(*jmps_ll_tail != NULL)
    {
      strncpy((*jmps_ll_tail)->name,name,SYMBOL_LEN);
      (*jmps_ll_tail)->name[SYMBOL_LEN] = 0;
      (*jmps_ll_tail)->line_no = line_no;
      (*jmps_ll_tail)->op = op;
      (*jmps_ll_tail)->bin_buf = bin_buf_curr;
      (*jmps_ll_tail)->bin_ptr = bin_buf_curr->size;
      (*jmps_ll_tail)->jmp_addr = jmp_addr;
      (*jmps_ll_tail)->type = type;
      (*jmps_ll_tail)->next = NULL;
      jmps_ll_tail = &((*jmps_ll_tail)->next);
    }
    else
      fatal_memory_error();
  }
}

/***************************************************************************
*
***************************************************************************/
int get_jmp(int op,char **s,int type)
{
  int term = 0,stat = 0;

  if(is_symbol(**s))
  {
    for(;;)
    {
      char c;

      c = *(*s + term);
      if((c == ' ') || (c == ',') || (c == 0))
        break;
      term++;
    }
    if(term > 0)
    {
      tsymbols_ll *p;
      char ts[LN_STR_LEN + 1];

      strncpy(ts,*s,term);
      ts[term] = 0;
      p = find_symbol(ts);
      if(p != NULL)
      {
        if(p->type & (SYMBOL_LABEL | SYMBOL_CODE | SYMBOL_EQU))
        {
          append_jmp(op,"",p->value,type);
          stat = 1;
        }
        else
          p_error("Symbol not a label");
      }
      else
      {
        append_jmp(op,ts,0,type);
        stat = 1;
      }
    }
  }
  else
  {
    int addr = 0;

    if(get_u16(s,&addr,0,0))
    {
      append_jmp(op,"",addr,type);
      stat = 1;
    }
  }
  if(stat)
    *s += term;
  return stat;
}

/***************************************************************************
*
***************************************************************************/
int label_search_ahead(char *lab)
{
  int l,stat = 0;

  l = ftell(fp);
  for(;;)
  {
    char s[LN_STR_LEN + 1];

    if(feof(fp))
      break;
    if(fgets(s,LN_STR_LEN,fp) != NULL)
    {
      if(!strncmp(s,lab,strlen(lab)))
      {
        stat = 1;
        break;
      }
    }
  }
  fseek(fp,l,SEEK_SET);
  return stat;
}

int get_symbol_val(char **s,int *val,int expected_type,int op,int type)
{
  int stat = 0,term = 0;

  *val = 0;
  if(is_symbol(**s))
  {
    for(;;)
    {
      char c;

      c = *(*s + term);
      if((c == ' ') || (c == ',') || (c == ')') || (c == 0))
        break;
      if(expected_type & SYMBOL_BIT)
        if(c == '.')
          break;
      term++;
    }
    if(term > 0)
    {
      tsymbols_ll *p;
      char ts[LN_STR_LEN + 1];

      strncpy(ts,*s,term);
      ts[term] = 0;
      p = find_symbol(ts);
      if(p != NULL)
      {
        if(p->type & expected_type)
        {
          stat = 1;
          *val = p->value;
          if((p->type == SYMBOL_DATA) && (expected_type & SYMBOL_BIT))
          {
            char *ps;

            ps = *s + term;
            if(*ps == '.')
            {
              ps++;
              term++;
              if((*ps >= '0') && (*ps <= '7'))
              {
                *val += *ps - '0';
                term++;
              }
              else
                p_error("Invalid bit specifier");
            }
            else
              p_error(un_opand_mess);
          }
        }
        else
          p_error("incompatable symbol");
      }
      else
      {
        if(op)
        {
          strcat(ts,":");
          if(label_search_ahead(ts))
          {
            char *pts;

            pts = ts;
            stat = 1;
            ts[strlen(ts) - 1] = 0;
            get_jmp(op,&pts,type);
            *val = 0;
          }
        }
      }
    }
  }
  if(stat)
    *s += term;
  return stat;
}

/***************************************************************************
*
***************************************************************************/
void get_def_symbol(char **s,char *sym)
{
  char fd[LN_STR_LEN + 1];

  void asym(int sym_t)
  {
    int val;

    skip_space(s);

    if(get_symbol_val(s,&val,
      sym_t == SYMBOL_BIT ? SYMBOL_BIT | SYMBOL_DATA : sym_t,0,0))
      append_symbol(sym,sym_t,val);
    else
      if(get_single_ordinal(s,&val))
        append_symbol(sym,sym_t,val);
  }

  skip_space(s);
  if(sscanf(*s,"%s",fd) == 1)
  {
    *s += strlen(fd);
    if(!strcmp(fd,"equ"))
      asym(SYMBOL_EQU);
    else if(!strcmp(fd,"="))
      asym(SYMBOL_EQU);
    else if(!strcmp(fd,"code"))
      asym(SYMBOL_CODE);
    else if(!strcmp(fd,"data"))
      asym(SYMBOL_DATA);
    else if(!strcmp(fd,"idata"))
      asym(SYMBOL_IDATA);
    else if(!strcmp(fd,"xdata"))
      asym(SYMBOL_XDATA);
    else if(!strcmp(fd,"bit"))
      asym(SYMBOL_BIT);
    else
      p_error(un_opor_mess);
  }
  else
    p_error(un_opor_mess);
}

/***************************************************************************
*
***************************************************************************/
void get_label(char **s)
{
  int x = 0,quote = 0;

  for(;;)
  {
    switch(*(*s + x))
    {
      case 0:
        return;
      case '\'':
        quote ^= 1;
        break;
      case ':':
        if(!quote)
        {
          skip_space(s);
          if(is_symbol(**s))
          {
            char ts[SYMBOL_LEN + 1];

            if(bin_buf_curr == NULL)
            {
              p_error("No ORG entry defined. "
                      "Must have an ORG statement before labels or code");
              exit(0);
            }

            x = 0;
            for(;;)
            {
              if(x > SYMBOL_LEN)
              {
                p_error(lab_too_long_mess);
                return;
              }
              if(**s == ':')
                break;
              if(((**s >= 'a') && (**s <= 'z')) || (**s == '_') ||
                 ((**s >= '0') && (**s <= '9')))
                ts[x] = **s;
              else
              {
                p_error("garbled label");
                return;
              }
              (*s)++;
              x++;
            }
            ts[x] = 0;
            (*s)++;
            append_symbol(
              ts,SYMBOL_LABEL,bin_buf_curr->size + bin_buf_curr->org);
          }
          return;
        }
        break;
    }
    x++;
  }
}

/***************************************************************************
* go through jmp list, and match up label addresses. update the binary
* image.
***************************************************************************/
void wirein_jmps(void)
{
  tjmps_ll *j;
  char *ofr_mess = "Jump out of range";

  void p_em(void)
  {
    error_count++;
    printf("%s:%d: error: %s\n",filename,j->line_no,ofr_mess);
  }

  int calc_sjmp(int offs)
  {
    int sjmp;

    sjmp = j->jmp_addr - (j->bin_ptr + j->bin_buf->org + offs);
    if((sjmp >= -128) && (sjmp <= 127))
      return sjmp;
    p_em();
    return 0;
  }

  void do_data(int i)
  {
    int x;

    x = j->jmp_addr;
    if(j->type == 2)
      x >>= 8;
    else if(j->type == 1)
      x &= 0xff;
    ammend_text8(j->bin_buf->buf,j->bin_ptr + i,x);
  }

  void wire(void)
  {
    switch(j->op)
    {
      case OP_ACALL:
      case OP_AJMP:
      {
        int op_j,osa;

        osa = (j->bin_ptr + j->bin_buf->org) & ~0x7ff;
        op_j = j->jmp_addr & ~0x7ff;

        if(osa != op_j)
          p_em();
        else
        {
          int x;

          op_j = j->jmp_addr & 0x7ff;
          x = (((op_j >> 3) & ~0x1f) | j->op) << 8;
          x |= op_j & 0xff;
          ammend_text16(j->bin_buf->buf,j->bin_ptr,(char *)&x);
        }
        break;
      }
      case OP_LCALL:
      case OP_LJMP:
      case OP_MOV_DPTR:
        ammend_text16(j->bin_buf->buf,j->bin_ptr + 1,(char *)&(j->jmp_addr));
        break;
      case OP_CJNE_ADIR:
      case OP_CJNE_ADATA:
      case OP_CJNE_REG:
      case OP_CJNE_PTR:
      case OP_DJNZ_DIR:
      case OP_JB_BIT:
      case OP_JNB_BIT:
      case OP_JBC_BIT:
        ammend_text8(j->bin_buf->buf,j->bin_ptr + 2,calc_sjmp(3));
        break;
      case OP_DJNZ_REG:
      case OP_JC:
      case OP_JNC:
      case OP_JNZ:
      case OP_JZ:
      case OP_SJMP:
        ammend_text8(j->bin_buf->buf,j->bin_ptr + 1,calc_sjmp(2));
        break;
      case OP_ADD_DATA:
      case OP_ADDC_DATA:
      case OP_MOV_A_DATA:
      case OP_MOV_REG_DATA:
      case OP_MOV_PTR_DATA:
      case OP_SUBB_DATA:
      case OP_ANL_ADATA:
      case OP_ORL_ADATA:
      case OP_XRL_ADATA:
        do_data(1);
        break;
      case OP_MOV_DIR_DATA:
      case OP_ANL_DIR2:
      case OP_ORL_DIR2:
      case OP_XRL_DIR2:
        do_data(2);
        break;
      default:
        error_count++;
        printf("%s:%d: error: Don't know how to wire label.\n",
          filename,j->line_no);
        break;
    }
  }

  j = jmps_ll;
  while(j != NULL)
  {
    if(j->name[0] == 0)
      wire();
    else
    {
      tsymbols_ll *l;

      l = find_symbol(j->name);
      if(l != NULL)
      {
        j->jmp_addr = l->value;
        wire();
      }
      else
      {
        error_count++;
        printf("%s:%d: error: Unsolvable label.\n",filename,j->line_no);
      }
    }
    j = j->next;
  }
}
